library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity gc_ipl is
	Port (  rom_adr : out std_logic_vector(20 downto 0);
		rom_dat : inout std_logic_vector(7 downto 0);
		rom_oe : out std_logic;
		rom_we : out std_logic;
		rom_ce : out std_logic;
		rom_reset : out std_logic;
--********************************************************
		-- aus chip sicht!
		uart_clock : in std_logic;
		uart_rx    : in std_logic;
		uart_tx    : out std_logic;
--********************************************************
		-- aus chip sicht!
		exi_in  : in std_logic;
		exi_out : out std_logic;
		exi_clk : in std_logic;
		exi_cs  : in std_logic;
		exi_out_ipl : in std_logic;
--********************************************************
		prog_exi_in : in std_logic;
		prog_exi_out : out std_logic;
		prog_exi_clk : in std_logic;
		prog_exi_cs : in std_logic;
		prog_exi_clk2 : in std_logic);
--********************************************************

end gc_ipl;

architecture behavioral of gc_ipl is

--********************************************************
signal rom_adr_host_sig: STD_LOGIC_VECTOR(20 downto 0);
signal rom_host_oe: std_logic;
signal rom_host_we: std_logic;
signal rom_host_data: std_logic_vector(7 downto 0);

signal exi_count  : std_logic_vector(5 downto 0);
signal exi_data: std_logic_vector(7 downto 0);
signal exi_out_latch: std_logic;
signal exi_loop_through: std_logic;

-- TX stuff: state is: 0 idle, 1 startbit, 2..10 data, 10 stop
signal uart_tx_state: std_logic_vector(3 downto 0);
signal uart_tx_counter: std_logic_vector(3 downto 0);
signal uart_tx_byte: std_logic_vector(7 downto 0);

signal uart_tx_start: std_logic;
signal uart_tx_done: std_logic;

-- RX stuff: state is: 0 idle, 1..9 data, 
signal uart_rx_state: std_logic_vector(3 downto 0);
signal uart_rx_counter: std_logic_vector(3 downto 0);
signal uart_rx_byte: std_logic_vector(7 downto 0);

signal rx_valid, tx_valid: std_logic;

-- protocol stuff:
signal rom_dat_ser: std_logic_vector(7 downto 0);
signal rom_adr_ser: std_logic_vector(19 downto 0);
signal rom_ser_oe, rom_ser_we: std_logic;
signal ser_state  : std_logic_vector(3 downto 0);

-- exi protocol
signal exi_address: std_logic_vector(25 downto 0);
signal config_bootorig: std_logic ;
signal config_bootprefix: std_logic_vector(3 downto 0);
signal do_rom: std_logic;
signal do_config: std_logic;
signal do_write: std_logic;
signal do_next_write: std_logic;
signal do_write_now: std_logic;
signal do_serial: std_logic;

signal real_exi_cs: std_logic;
signal real_exi_in: std_logic;
signal real_exi_clk : std_logic;
signal real_exi_out: std_logic;
signal prog_exi_clk_state: std_logic;

signal test_toggle: std_logic_vector(3 downto 0);

begin



--************************************************************************************************
--  		Global Process
--
-- Test for program or normal mode, and setup signal routes
-- and other signals for it
--************************************************************************************************
Switch_Global: process	(	rom_adr_host_sig,
				rom_dat, uart_clock,
				rom_ser_oe, rom_ser_we, rom_adr_ser, rom_dat_ser, ser_state,
				rom_host_we, rom_host_data, rom_host_oe
			)
begin


	rom_reset <= '1';

	rom_ce <= '0';
	rom_adr <= "0" & rom_adr_ser;
	if ser_state /= x"1" and ser_state /= x"6" and ser_state /= x"7" then
		rom_dat <= (others => 'Z');
	else
		rom_dat <= rom_dat_ser;
	end if;
	rom_oe <= rom_ser_oe;
	rom_we <= rom_ser_we;

	if (uart_clock'event and uart_clock = '1') then
		test_toggle <= test_toggle + '1';
		if rx_valid = '1' then
			case uart_rx_byte(7 downto 4) is
				when x"0" =>
					rom_dat_ser(3 downto 0) <= uart_rx_byte(3 downto 0);
				when x"1" =>
					rom_dat_ser(7 downto 4) <= uart_rx_byte(3 downto 0);
					ser_state <= x"1";  -- write cycle
				when x"2" =>
					case uart_rx_byte(3 downto 0) is
						when x"0" =>
							ser_state <= x"2";  -- read cycle w/o postinc
						when x"1" =>
							ser_state <= x"3";  -- read cycle with postinc
						when x"2" =>
							ser_state <= x"4";  -- exi
						when others =>
							ser_state <= x"5";  -- illegal
					end case;
				when x"3" =>
					rom_adr_ser(3 downto 0) <= uart_rx_byte(3 downto 0);
				when x"4" =>
					rom_adr_ser(7 downto 4) <= uart_rx_byte(3 downto 0);
				when x"5" =>
					rom_adr_ser(11 downto 8) <= uart_rx_byte(3 downto 0);
				when x"6" =>
					rom_adr_ser(15 downto 12) <= uart_rx_byte(3 downto 0);
				when x"7" =>
					rom_adr_ser(19 downto 16) <= uart_rx_byte(3 downto 0);
				when others =>
					ser_state <= x"5";
			end case;
		else
			case ser_state is
				when x"0" => -- idle
					rom_ser_we <= '1';
					rom_ser_oe <= '1';
					if uart_tx_done = '1' then
						uart_tx_start <= '0';
					end if;
--					tx_valid <= '0';
				when x"1" => -- write cycle
					rom_ser_we <= '0';
					rom_ser_oe <= '1';
					ser_state <= x"6";
				when x"2" => -- read w/o postinc
					rom_ser_oe <= '0';
					ser_state <= x"8";
				when x"3" => -- read w postinc
					rom_ser_oe <= '0';
					ser_state <= x"9";
				when x"6" =>
					rom_ser_we <= '1';
					rom_ser_oe <= '1';
					ser_state <= x"7";
				when x"7" =>
					rom_ser_we <= '1';
					rom_ser_oe <= '1';
					ser_state <= x"0";
				when x"8" =>
					uart_tx_byte <= rom_dat;
--					tx_valid <= '1';
					uart_tx_start <= '1';
					ser_state <= x"0";
				when x"9" =>
					uart_tx_byte <= rom_dat;
--					tx_valid <= '1';
					uart_tx_start <= '1';
					rom_adr_ser(3 downto 0) <= rom_adr_ser(3 downto 0) + '1';
					ser_state <= x"0";
				when others =>
					ser_state <= x"0";
					uart_tx_byte <= x"40";
--					tx_valid <= '1';
					uart_tx_start <= '1';
			end case;
		end if;
	end if;
end process;

UART_TX_process: process (uart_clock, uart_tx_state, uart_tx_start)
begin
	if (uart_clock'event and uart_clock = '1') then
		if uart_tx_state = x"0" then
			if uart_tx_start = '1' and uart_tx_done = '0' then
				uart_tx_state <= x"1";
			end if;

			if uart_tx_start = '0' then
				uart_tx_done <= '0';
			end if;

			uart_tx <= '1';
			uart_tx_counter <= (others => '0');
		else
			if (uart_tx_counter = x"0") then
				if uart_tx_state = x"1" then
					uart_tx <= '0'; uart_tx_state <= x"2"; -- start
				elsif (uart_tx_state >= x"2") and (uart_tx_state <= x"A") then
					if uart_tx_state = x"A" then
						uart_tx <= '1';
					else
						uart_tx <= uart_tx_byte(conv_integer(uart_tx_state) - 2);
					end if;
					uart_tx_state <= uart_tx_state + '1';
				else
					uart_tx_state <= x"0";
					uart_tx_done <= '1';
					uart_tx <= '1';
				end if;
			end if;
			uart_tx_counter <= uart_tx_counter + '1';
		end if;
	end if;
end process;

-- uart_tx <= test_toggle(3);

UART_RX_process: process (uart_clock, uart_rx_state)
begin
	if (uart_clock'event and uart_clock = '1') then
		if uart_rx_state = x"0" then  -- wait for startbit
			if uart_rx = '1' then
				uart_rx_state <= x"0";
			else
				uart_rx_state <= x"1";
			end if;
			uart_rx_counter <= (others => '0');
			rx_valid <= '0'; 
		else
			if (uart_rx_counter = "1000") then  -- sample in the middle
				if uart_rx_state >= x"2" and uart_rx_state <= x"9" then
					uart_rx_byte <= uart_rx & uart_rx_byte(7 downto 1);
				end if;

				if uart_rx_state = x"A" then  -- stopbit reached?
					uart_rx_state <= x"0"; -- go into idle
					rx_valid <= '1';
				else
					uart_rx_state <= uart_rx_state + '1';
				end if;
			end if;
			uart_rx_counter <= uart_rx_counter + '1';
		end if;
	end if;
end process;



end behavioral;



